//**************************************************************************
//*                     This file is part of the                           *
//*                 Mpxplay/MMC - multimedia player.                       *
//*                  The source code of Mpxplay is                         *
//*        (C) copyright 1998-2019 by PDSoft (Attila Padar)                *
//*                http://mpxplay.sourceforge.net                          *
//*                  email: mpxplay@freemail.hu                            *
//**************************************************************************
//*  This program is distributed in the hope that it will be useful,       *
//*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *
//*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                  *
//*  Please contact with the author (with me) if you want to use           *
//*  or modify this source.                                                *
//**************************************************************************
//function: FFMPEG decoder common functions and callback

#include "ffmpgdec.h"

#ifdef MPXPLAY_LINK_INFILE_FF_MPEG

#include "mpxplay.h"
#include "tagging.h"

#ifdef MPXPLAY_GUI_QT
extern unsigned long mpxplay_config_videoplayer_control;
#endif

//------------------------------------------------------------------------------------------------------------------
void in_ffmpgdec_avsync_assign(struct ffmpg_demuxer_data_s *ffmpi, mpxp_int64_t pts)
{
	ffmpi->avsync_timestamp = pts;
	ffmpi->avsync_clocktime = pds_gettimeu();
}

// convert stream time based dts/pts to AV_TIME_BASEd
mpxp_int64_t in_ffmpgdec_get_avpacket_ts_timestamp(struct ffmpg_demuxer_data_s *ffmpi, unsigned int streamtype_index, mpxp_int64_t pkt_ts)
{
	AVStream *stream = (ffmpi && (streamtype_index < MPXPLAY_STREAMTYPEINDEX_NUM))? ffmpi->selected_avstream[streamtype_index] : NULL;

	if(stream && stream->time_base.num && stream->time_base.den)
		pkt_ts = pkt_ts * AV_TIME_BASE * stream->time_base.num / stream->time_base.den;
	else
		pkt_ts = pkt_ts * AV_TIME_BASE / 90000.0; // !!! MPEG like pts to usecs

	pkt_ts += INFFMPG_TIMESTAMP_SHIFT;  // first timestamp can be 0 (-audio_delay gives negative sync)

	return pkt_ts; // in AV_TIME_BASE
}

// get AV_TIME_BASE based timestamp for an AVPacket
mpxp_int64_t in_ffmpgdec_get_avpacket_timestamp(struct ffmpg_demuxer_data_s *ffmpi, unsigned int streamtype_index, AVPacket *pkt)
{
	mpxp_int64_t pts = ((pkt->pts >= 0) && ((pkt->pts >= pkt->dts) || (pkt->dts < 0)))? pkt->pts : pkt->dts;
	return in_ffmpgdec_get_avpacket_ts_timestamp(ffmpi, streamtype_index, pts);
}

// select pts/dts (which seems to be the better)
mpxp_int64_t in_ffmpgdec_select_avframe_pts(AVFrame *avframe)
{
	mpxp_int64_t pts = (avframe->pts >= 0)? avframe->pts : (((avframe->pkt_dts >= 0) && ((avframe->pkt_dts <= avframe->pkt_pts) || (avframe->pkt_pts < 0)))? avframe->pkt_dts : avframe->pkt_pts);
	return pts;
}

// get AV_TIME_BASE based timestamp for an AVFrame
mpxp_int64_t in_ffmpgdec_get_avframe_timestamp(struct ffmpg_demuxer_data_s *ffmpi, unsigned int streamtype_index, AVFrame *avframe)
{
	mpxp_int64_t pts = in_ffmpgdec_select_avframe_pts(avframe);
	if(pts < 0)
		return 0;
	return in_ffmpgdec_get_avpacket_ts_timestamp(ffmpi, streamtype_index, pts);
}

static mpxp_int64_t inffmpg_get_avframe_pts_corrected(struct ffmpg_demuxer_data_s *ffmpi, unsigned int streamtype_index, AVFrame *avframe)
{
	mpxp_int64_t dts = in_ffmpgdec_get_avframe_timestamp(ffmpi, streamtype_index, avframe);
#ifdef INFFMPG_USE_VIDEOPTS_CORRECTON
	mpxp_int64_t dts_video_last = ffmpi->stream_last_pts[MPXPLAY_STREAMTYPEINDEX_VIDEO];

	if((streamtype_index == MPXPLAY_STREAMTYPEINDEX_VIDEO) && (dts_video_last > INFFMPG_TIMESTAMP_SHIFT))  // TODO: wide test for non-streams too
	{
		AVStream *st = ffmpi->selected_avstream[streamtype_index];
		mpxp_int64_t dts_file = dts;
		mpxp_int64_t avduration = av_frame_get_pkt_duration(avframe);
		mpxp_int64_t duration_us = (st && st->time_base.num && st->time_base.den)? (avduration * AV_TIME_BASE * st->time_base.num / st->time_base.den) : (avduration * AV_TIME_BASE);

		if(((duration_us <= 1000) || (duration_us >= (AV_TIME_BASE / 2))) && st && st->avg_frame_rate.num && st->avg_frame_rate.den)  // TODO: max 999fps / 2fps is allowed
			duration_us = AV_TIME_BASE * st->avg_frame_rate.den / st->avg_frame_rate.num;
		if(duration_us <= 1000) // TODO: max 999 fps (because some MKVs contain invalid 1000 fps value)
			duration_us = 40000; // TODO: fake 25 fps

		if((dts <= INFFMPG_TIMESTAMP_SHIFT) || (dts <= dts_video_last) || (dts > (dts_video_last + duration_us * 2))) // TODO: more tests for this workaround if video pts is not sequential
		{
			mpxp_int64_t dts_audio = ffmpi->stream_last_pts[MPXPLAY_STREAMTYPEINDEX_AUDIO];
			unsigned int seq_corr = 1;

			if( (dts > INFFMPG_TIMESTAMP_SHIFT)
			 && ((dts < (dts_video_last - INFFMPG_TIMESTAMP_MAX_AVDIFFERENCE)) || (dts > (dts_video_last + INFFMPG_TIMESTAMP_MAX_AVDIFFERENCE)))
			 && (dts_audio > INFFMPG_TIMESTAMP_SHIFT)
			){
				mpxp_int64_t dts_adiff = dts_video_last - dts_audio;

				if(dts_adiff < 0)
					dts_adiff = -dts_adiff;

				if(dts_adiff > INFFMPG_TIMESTAMP_MAX_AVDIFFERENCE) // TODO: more test for this workaround if video dts/pts is completely false
				{
					dts = dts_audio;
					seq_corr = 0;
					mpxplay_debugf(MPXPLAY_DEBUG_QUEUEGET,"get_avpacket_dts_corrected AVSYNC dts:%lld diff:%lld adi:%llu cd:%llu cp:%llu dur:%llu",
						dts, (dts - dts_video_last), dts_adiff, (avframe->pkt_dts), (avframe->pkt_pts), duration_us);
				}
				else
				{
//					mpxplay_debugf(MPXPLAY_DEBUG_QUEUEGET,"get_avpacket_dts_corrected NODIFF dts:%llu diff:%lld vdts:%llu vpts:%llu",
//							dts, (dts - dts_video_last),
//							((avpkt)? avpkt->dts : avframe->pkt_dts), ((avpkt)? avpkt->pts : avframe->pkt_pts));
				}
			}

			if(seq_corr){
				mpxp_int64_t correction = duration_us >> 4; // a smaller correction to file/frame dts from calculated
				mpxp_int64_t dvl = dts_video_last + duration_us;
				dts = dvl + ((dts <= dts_video_last)? -correction : correction);
			}
			mpxplay_debugf(MPXPLAY_DEBUG_QUEUEGET,"BADSEQ dr:%llu dts:%lld dtsf:%lld last:%lld diff:%lld df:%lld da:%lld dtsa:%lld pdts:%lld ppts:%lld",
				duration_us, dts, dts_file, dts_video_last, (dts - dts_video_last),
				(dts_file - dts_video_last), (dts_video_last - dts_audio), dts_audio, (avframe->pkt_dts), (avframe->pkt_pts));
		}
	}
#endif
	return dts; // in AV_TIME_BASE (usec)
}

//------------------------------------------------------------------------------------------------------------------
#ifdef MPXPLAY_GUI_QT

#define INFFMPG_STREAMINFO_MAX_LEN 128

static struct mpxplay_programinfo_t *inffmpg_programinfolist_fill(struct ffmpg_demuxer_data_s *ffmpi)
{
	struct mpxplay_programinfo_t *program_list;
	int i, pc;

	if(!in_ffmpgdec_content_has_programs(ffmpi))
		return NULL;

	program_list = pds_calloc(1, sizeof(*program_list));
	if(!program_list)
		return program_list;

	for(i = 0; i < ffmpi->fctx->nb_programs; i++)
	{
		AVProgram *prg = ffmpi->fctx->programs[i];
		if(prg)
		{
			AVDictionaryEntry *t = av_dict_get(prg->metadata, "service_name", NULL, 0); // TODO: check, program elements without service names are ignored!
			if(t && t->value)
				program_list->nb_programs++;  // number of valid programs
		}
	}

	if(program_list->nb_programs <= 1)
		goto err_out_ipf;

	program_list->program_names = pds_calloc(program_list->nb_programs, sizeof(program_list->program_names[0]));
	program_list->program_ids = pds_calloc(program_list->nb_programs, sizeof(program_list->program_ids[0]));
	if(!program_list->program_names || !program_list->program_ids)
		goto err_out_ipf;

	for(i = 0, pc = 0; (i < ffmpi->fctx->nb_programs) && (pc < program_list->nb_programs); i++)
	{
		AVProgram *prg = ffmpi->fctx->programs[i];
		AVDictionaryEntry *t;

		if(!prg)
			continue;

		t = av_dict_get(prg->metadata, "service_name", NULL, 0);
		if(!t || !t->value)
			continue;

		if(prg->id == ffmpi->program_number_current)
			program_list->selected_prognum = pc;
		program_list->program_ids[pc] = prg->id;
		program_list->program_names[pc] = malloc(pds_strlen(t->value) + 1);
		if(program_list->program_names[pc])
			pds_strcpy(program_list->program_names[pc], t->value);
		pc++;
	}

	return program_list;

err_out_ipf:
	if(program_list){
		pds_free(program_list->program_names);
		pds_free(program_list->program_ids);
		pds_free(program_list);
	}
	return NULL;
}

static struct mpxplay_streaminfo_t *inffmpg_streaminfolist_fill(struct ffmpg_demuxer_data_s *ffmpi)
{
	struct mpxplay_streaminfo_t *stream_list = pds_calloc(1, sizeof(*stream_list));
	int i, j;
	if(!stream_list)
		return stream_list;
	for(i = 0; i <= MPXPLAY_STREAMTYPEINDEX_SUBTITLE; i++)
	{
		unsigned int nb_in_streams, nb_ext_streams, nb_all_streams;
		int stream_number = 0, stream_ff_idx = INFFMPG_STREAM_INDEX_DISABLE;
		in_ffmpstrm_streams_count(ffmpi, i, &stream_number, &stream_ff_idx, NULL);
		nb_in_streams = ffmpi->nb_streams[i];
		nb_ext_streams = (i != MPXPLAY_STREAMTYPEINDEX_VIDEO)? ffmpi->external_file_infos[i].nb_external_files : 0; // FIXME: incorrect extra video stream list
		nb_all_streams = nb_in_streams + nb_ext_streams;
		mpxplay_debugf(MPXPLAY_DEBUG_STRMLIST, "inffmpg_streams %d ni:%d ne:%d", i, nb_in_streams, nb_ext_streams);
		if(nb_all_streams)
		{
			struct ffmpeg_external_filelist_elem_s *external_files_chain;
			if(i == MPXPLAY_STREAMTYPEINDEX_SUBTITLE)
			{
				nb_in_streams++; nb_all_streams++; // "Disable" selection
			}
			stream_list->stream_names[i] = pds_calloc(nb_all_streams, sizeof(stream_list->stream_names[i]));
			stream_list->stream_idx[i] = pds_calloc(nb_all_streams, sizeof(stream_list->stream_idx[i]));
			if(stream_list->stream_names[i] && stream_list->stream_idx[i])
			{
				int j = 0;
				stream_list->nb_entries[i] = nb_all_streams;
				stream_list->selected_stream_index[i] = (ffmpi->selected_stream_number[i] < INFFMPG_STREAM_INDEX_DISABLE)? INFFMPG_STREAM_INDEX_DISABLE : ffmpi->selected_stream_number[i];

				if(i == MPXPLAY_STREAMTYPEINDEX_SUBTITLE)
				{
					stream_list->stream_names[i][0] = pds_malloc(INFFMPG_STREAMINFO_MAX_LEN);
					pds_strcpy(stream_list->stream_names[i][0], "S:  Disable");
					stream_list->stream_idx[i][0] = INFFMPG_STREAM_INDEX_DISABLE;
					j++;
				}

				// internal stream list (from the primary file)
				stream_number = 0;
				for( ; (j < nb_in_streams); j++, stream_number++)
				{
					int sn_save;
					AVStream *st;
					AVDictionaryEntry *t;
					char language_str[32] = "", bit_rate_str[32] = "", channel_str[32];

					sn_save = stream_number;
					stream_ff_idx = INFFMPG_STREAM_INDEX_DISABLE;
					if(in_ffmpstrm_streams_count(ffmpi, i, &stream_number, &stream_ff_idx, NULL) <= 0)
					{
						mpxplay_debugf(MPXPLAY_DEBUG_WARNING, "inffmpg_streams_count FAILED %d", i);
						break;
					}
					if((sn_save > stream_number) || (stream_ff_idx < 0) || (stream_ff_idx >= ffmpi->fctx->nb_streams))
					{
						mpxplay_debugf(MPXPLAY_DEBUG_WARNING, "inffmpg_streams_count stream_ff_idx FAILED i:%d sn:%d idx:%d", i, stream_number, stream_ff_idx);
						break;
					}

					st = ffmpi->fctx->streams[stream_ff_idx];
					if(!st)
						continue;

					stream_list->stream_names[i][j] = pds_malloc(INFFMPG_STREAMINFO_MAX_LEN);
					if(!stream_list->stream_names[i][j])
						break;

					t = av_dict_get(st->metadata, "language", NULL, 0);

					//mpxplay_debugf(MPXPLAY_DEBUG_STRMLIST, "inffmpg_streaminfolist_fill i:%d ffi:%d", i, stream_ff_idx);

					if(t && t->value && t->value[0] && ((i == MPXPLAY_STREAMTYPEINDEX_SUBTITLE) || (pds_stricmp(t->value,"und") != 0)))
					{
						const char *language_name = t->value;
						if(!funcbit_test(mpxplay_config_videoplayer_control, MPXPLAY_CONFIG_VIDEOPLAYERCONTROL_STREAMLIST_SHORT_LANGNAMES))
						{
							language_name = mpxplay_ffmplang_getlangname_by_langcode(language_name);
							if(!language_name)
								language_name = t->value;
						}
						snprintf(language_str, sizeof(language_str), " [%s] ", language_name);
					}

					if(st->codecpar)
					{
						int bit_rate = ((st->codecpar->bit_rate)? st->codecpar->bit_rate : st->codec->bit_rate) / 1000, channel_num;
						if(bit_rate > 0)
							snprintf(bit_rate_str, sizeof(bit_rate_str), ", %d kb/s", bit_rate);
						switch(i)
						{
							case MPXPLAY_STREAMTYPEINDEX_AUDIO:
								channel_num = st->codecpar->channels;
								switch(channel_num){
									case 1: pds_strcpy(channel_str, "Mono"); break;
									case 2: pds_strcpy(channel_str, "Stereo"); break;
									case 6: pds_strcpy(channel_str, "5.1 chans"); break;
									default: snprintf(channel_str, sizeof(channel_str), "%d chans", channel_num); break;
								}
								snprintf(stream_list->stream_names[i][j], INFFMPG_STREAMINFO_MAX_LEN, "A: %s%s, %d Hz, %s%s",
									language_str, avcodec_get_name(st->codecpar->codec_id), st->codecpar->sample_rate, channel_str, bit_rate_str);
								break;
							case MPXPLAY_STREAMTYPEINDEX_VIDEO:
								snprintf(stream_list->stream_names[i][j], INFFMPG_STREAMINFO_MAX_LEN, "V: %s%s, %dx%d, %2.2f fps%s",
									language_str, avcodec_get_name(st->codecpar->codec_id),
									st->codecpar->width, st->codecpar->height,
									(st->avg_frame_rate.num && st->avg_frame_rate.den)? (float)(float)st->avg_frame_rate.num / (float)st->avg_frame_rate.den : (float)0,
									bit_rate_str);
								break;
							case MPXPLAY_STREAMTYPEINDEX_SUBTITLE:
								snprintf(stream_list->stream_names[i][j], INFFMPG_STREAMINFO_MAX_LEN, "S: %s", language_str);
								break;
						}
					}
					else
					{
						char *shorttypename = (i == MPXPLAY_STREAMTYPEINDEX_AUDIO)? "A:" : ((i == MPXPLAY_STREAMTYPEINDEX_VIDEO)? "V:" : "S:");
						snprintf(stream_list->stream_names[i][j], INFFMPG_STREAMINFO_MAX_LEN, "%s Track %d%s", shorttypename, (j + 1), language_str); // TODO: better stream names
					}
					stream_list->stream_idx[i][j] = stream_number;
					mpxplay_debugf(MPXPLAY_DEBUG_STRMLIST, "MPXPLAY_INFILE_CBKCTRL_GET_STREAMLIST %d/%d %d/%d %s", i, j, stream_number, stream_ff_idx, stream_list->stream_names[i][j]);
				}

				// external stream list (from extra audio/subtitle stream files)
				external_files_chain = ffmpi->external_file_infos[i].external_files_chain;
				for( ; (j < nb_all_streams) && external_files_chain; j++)
				{
					char *shorttypename, language_str[32] = "", bit_rate_str[32] = "", channel_str[32] ="", samplerate_str[32] = "";
					int bit_rate;

					stream_list->stream_names[i][j] = pds_malloc(INFFMPG_STREAMINFO_MAX_LEN);
					if(!stream_list->stream_names[i][j])
						break;

					if(external_files_chain->language_name[0] || external_files_chain->language_code[0])
						snprintf(language_str, sizeof(language_str), " [%s] ", ((!funcbit_test(mpxplay_config_videoplayer_control, MPXPLAY_CONFIG_VIDEOPLAYERCONTROL_STREAMLIST_SHORT_LANGNAMES) && external_files_chain->language_name[0])? external_files_chain->language_name : external_files_chain->language_code));

					bit_rate = external_files_chain->bitrate / 1000;
					if(bit_rate > 0)
						snprintf(bit_rate_str, sizeof(bit_rate_str), ", %d kb/s", bit_rate);

					if(external_files_chain->channels > 0)
					{
						switch(external_files_chain->channels){
							case 1: pds_strcpy(channel_str, ", Mono"); break;
							case 2: pds_strcpy(channel_str, ", Stereo"); break;
							case 6: pds_strcpy(channel_str, ", 5.1 chans"); break;
							default: snprintf(channel_str, sizeof(channel_str), ", %d chans", external_files_chain->channels); break;
						}
					}

					if(external_files_chain->samplerate > 0)
						snprintf(samplerate_str, sizeof(samplerate_str), ", %d Hz", external_files_chain->samplerate);

					shorttypename = (i == MPXPLAY_STREAMTYPEINDEX_AUDIO)? "A:" : ((i == MPXPLAY_STREAMTYPEINDEX_VIDEO)? "V:" : "S:");
					snprintf(stream_list->stream_names[i][j], INFFMPG_STREAMINFO_MAX_LEN, "%s %s%s%s%s%s", shorttypename, language_str,
							((external_files_chain->format_name[0])? external_files_chain->format_name : ((external_files_chain->codec_id > AV_CODEC_ID_NONE)? avcodec_get_name(external_files_chain->codec_id) : pds_getfilename_from_fullname(external_files_chain->external_filename))),
							samplerate_str, channel_str, bit_rate_str);

					stream_list->stream_idx[i][j] = external_files_chain->stream_index;

					external_files_chain = external_files_chain->next_external_file;
				}

				if(j <= ((i == MPXPLAY_STREAMTYPEINDEX_SUBTITLE)? 1 : 0))  // number of streams has changed or no valid stream found
				{
					stream_list->nb_entries[i] = 0;
					stream_list->selected_stream_index[i] = INFFMPG_STREAM_INDEX_DISABLE;
					for(j = 0; j < nb_all_streams; j++)
						if(stream_list->stream_names[i][j])
							pds_free(stream_list->stream_names[i][j]);
					pds_free(stream_list->stream_names[i]);
					pds_free(stream_list->stream_idx[i]);
					stream_list->stream_names[i] = NULL;
					stream_list->stream_idx[i] = NULL;
				}
			}
		}
	}
	return stream_list;
}

static void inffmpg_pictureinfo_metadata_update(struct ffmpg_demuxer_data_s *ffmpi, AVDictionary *metadata)
{
	if((ffmpi->flags & INFFMPG_FLAG_STILLPICTURE) && ffmpi->miis)
	{
		AVDictionaryEntry *e_make = av_dict_get(metadata, "Make", NULL, 0);
		AVDictionaryEntry *e_model = av_dict_get(metadata, "Model", NULL, 0);
		AVDictionaryEntry *e_fnum = av_dict_get(metadata, "FNumber", NULL, 0);
		AVDictionaryEntry *e_etime = av_dict_get(metadata, "ExposureTime", NULL, 0);
		AVDictionaryEntry *e_iso = av_dict_get(metadata, "ISOSpeedRatings", NULL, 0);
		int len = -1, val1, val2;
		float f_fnum = 0.0f, f_etime = 0.0f;
		char str[256];

		if(((!e_make || !e_make->value) && (!e_model || !e_model->value)) || !e_fnum || !e_etime || !e_fnum->value || !e_etime->value)
			return;

		if(e_fnum){
			val1 = val2 = 1;
			sscanf(e_fnum->value, "%d:%d", &val1, &val2);
			if(val2 > 0)
				f_fnum = (float) val1 / (float)val2;
		}
		if(e_etime){
			val1 = val2 = 1;
			sscanf(e_etime->value, "%d:%d", &val1, &val2);
			if(val2 > 0)
				f_etime = 1.0f / ((float) val1 / (float)val2);
		}
		snprintf(str, sizeof(str), "%s %s, F%.1f, 1/%.0f, ISO%d", ((e_make)? e_make->value : ""), ((e_model)? e_model->value : ""), f_fnum, f_etime, ((e_iso)? pds_atol(e_iso->value) : 0));
		ffmpi->miis->control_cb(ffmpi->fbds,MPXPLAY_CFGFUNCNUM_INFILE_ENTRY_TAG_PACKFUNCTTI3I(MPXPLAY_CFGFUNCNUM_INFILE_ENTRY_TAG_PUT,MPXPLAY_TEXTCONV_TYPE_UTF8,I3I_COMMENT),str,&len);
	}
}

#endif // MPXPLAY_GUI_QT

//--------------------------------------------------------------------------------------------------------------------------------------
// HW based video decoding (DXVA and D3D11VA hardware accelerators are enabled in FFmpeg lib)

static int in_ffmpgdec_hwdec_init(struct ffmpg_demuxer_data_s *ffmpi, AVCodecContext *codec_ctx_new, const AVCodec *codec_av_new)
{
	int dxva_result = -1;

#ifdef MPXPLAY_LINK_DXVA
	if(funcbit_test(mpxplay_config_videoplayer_control, MPXPLAY_CONFIG_VIDEOPLAYERCONTROL_USEHWDECODING))
	{
		if(codec_ctx_new->coded_width <= 0)
			codec_ctx_new->coded_width = codec_ctx_new->width;
		if(codec_ctx_new->coded_height <= 0)
			codec_ctx_new->coded_height = codec_ctx_new->height;
		if((codec_ctx_new->coded_width > 0) && (codec_ctx_new->coded_height > 0)
		 && ( !funcbit_test(mpxplay_config_videoplayer_control, MPXPLAY_CONFIG_VIDEOPLAYERCONTROL_DISABLESDHWDEC)
			  || (codec_ctx_new->coded_width > 720) || (codec_ctx_new->coded_height > 540))
		){
			enum AVHWDeviceType hw_dev_type = AV_HWDEVICE_TYPE_NONE;
			int i;
			for(i = 0;; i++) {
				const AVCodecHWConfig *config = avcodec_get_hw_config(codec_av_new, i);
				if (!config)
					return dxva_result;
				if( (config->methods & AV_CODEC_HW_CONFIG_METHOD_HW_DEVICE_CTX)
				 && ((config->device_type == AV_HWDEVICE_TYPE_DXVA2) || (config->device_type == AV_HWDEVICE_TYPE_D3D11VA)) // !!! (valid in windows only)
				){
					hw_dev_type = config->device_type;
					ffmpi->hw_pix_fmt = config->pix_fmt; // TODO: should be used in video_worker_videodecoder_frame_process
					break;
				}
			}
			dxva_result = av_hwdevice_ctx_create(&ffmpi->hw_device_ctx, hw_dev_type, NULL, NULL, 0);
			if(dxva_result >= 0)
			{
				codec_ctx_new->hw_device_ctx = av_buffer_ref(ffmpi->hw_device_ctx);
				if(!codec_ctx_new->hw_device_ctx)
					goto err_out_dxva;
				codec_ctx_new->opaque = (void *)ffmpi; // !!!
			}
		}
	}
#endif

	return dxva_result;

#ifdef MPXPLAY_LINK_DXVA
err_out_dxva:
	av_buffer_unref(&ffmpi->hw_device_ctx);
	ffmpi->hw_pix_fmt = AV_PIX_FMT_NONE;
	codec_ctx_new->opaque = NULL;
	return -1;
#endif
}

void in_ffmpgdec_hwdec_deinit(AVCodecContext *codec_ctx)
{
#ifdef MPXPLAY_LINK_DXVA
	if(codec_ctx->opaque)  // !!! assumed ffmpg_demuxer_data_s
	{
		struct ffmpg_demuxer_data_s *ffmpi = (struct ffmpg_demuxer_data_s *)codec_ctx->opaque;
		av_buffer_unref(&ffmpi->hw_device_ctx);
	}
#endif
}

//--------------------------------------------------------------------------------------------------------------------------------------
// number of software (not HW) based FFmpeg decoder threads
static int inffmpgdec_get_nb_sw_video_decoder_threads(struct ffmpg_demuxer_data_s *ffmpi, AVCodecContext *codec_ctx)
{
	int thread_count = 1;
	if(!(ffmpi->flags & INFFMPG_FLAG_STILLPICTURE))
	{
		thread_count = pds_threads_get_number_of_processors();
		if(thread_count < 2)
			thread_count = 2;
	}
	return thread_count;
}

static int inffmpgdec_get_avtimesync_us(struct ffmpg_demuxer_data_s *ffmpi, unsigned int stream_type, mpxp_int64_t *frame_timestamp_us)
{
	int retcode = MPXPLAY_ERROR_INFILE_NODATA;

	if(stream_type >= MPXPLAY_STREAMTYPEINDEX_NUM)
		return retcode;

	if(ffmpi->avsync_timestamp < 0)
	{
		if(funcbit_test(ffmpi->flags, INFFMPG_FLAG_INDEPENDENTSTREAMSYNC))
		{
			struct ffmpeg_external_files_info_s *extfilehandler =  &ffmpi->external_file_infos[stream_type];
			if((extfilehandler->avclocktime_us < 0) || (extfilehandler->avtimestamp_us < 0))
			{
				extfilehandler->avclocktime_us = pds_gettimeu();
				extfilehandler->avtimestamp_us = *frame_timestamp_us; // !!! frame timestamp must be passed
			}
			*frame_timestamp_us = pds_gettimeu() - extfilehandler->avclocktime_us + extfilehandler->avtimestamp_us;
			retcode = MPXPLAY_ERROR_INFILE_OK;
		}
		else if(!ffmpi->codec_ctx[MPXPLAY_STREAMTYPEINDEX_AUDIO] || (ffmpi->packet_queues[MPXPLAY_STREAMTYPEINDEX_VIDEO].nb_packets >= INFFMPG_QUEUE_PACKETNUM_MAX))
		{
			in_ffmpgdec_avsync_assign(ffmpi, ffmpi->stream_last_pts[MPXPLAY_STREAMTYPEINDEX_VIDEO]);
			mpxplay_debugf(MPXPLAY_DEBUG_TIMESTAMP, "in_ffmpgdec_infile_callback TIMESYNC: %llu", ffmpi->avsync_timestamp);
		}
		else
		{
			*frame_timestamp_us = 0;
		}
	}

	if(ffmpi->avsync_timestamp >= 0)
	{
		*frame_timestamp_us = pds_gettimeu() - ffmpi->avsync_clocktime + ffmpi->avsync_timestamp;
		retcode = MPXPLAY_ERROR_INFILE_OK;
		//mpxplay_debugf(MPXPLAY_DEBUG_TIMESTAMP, "in_ffmpgdec_infile_callback TIMESTAMP: %llu vts:%llu", *frame_timestamp_us, ffmpi->stream_last_pts[MPXPLAY_STREAMTYPEINDEX_VIDEO]);
	}

	return retcode;
}

static int inffmpgdec_codecctx_init(struct ffmpg_demuxer_data_s *ffmpi, unsigned int stream_type, mpxplay_packetlist_t **packet_dest)
{
	int retcode = MPXPLAY_ERROR_INFILE_NODATA;
	mpxplay_packetlist_t *pktlist_elem;

	if(stream_type >= MPXPLAY_STREAMTYPEINDEX_NUM)
		return retcode;

	//mpxplay_debugf(MPXPLAY_DEBUG_CALLBACK, "MPXPLAY_INFILE_CBKCTRL_GET_CODECCTXINIT_FFMPEG %d", stream_type);
	pktlist_elem = NULL;
	mpxplay_ffmpgdec_packetqueue_get((void *)ffmpi, &ffmpi->codecctx_queues[stream_type], &pktlist_elem, 0);
	if(pktlist_elem){
		if((pktlist_elem->flags & MPXPLAY_PACKETLISTFLAG_INIT) && pktlist_elem->codec_ctx){
			AVCodecContext *codec_ctx_new = (AVCodecContext *)pktlist_elem->codec_ctx;
			const AVCodec *codec_av_new = codec_ctx_new->codec;
			int dxva_result = -1;
			if(stream_type == MPXPLAY_STREAMTYPEINDEX_VIDEO){
				dxva_result = in_ffmpgdec_hwdec_init(ffmpi, codec_ctx_new, codec_av_new);
				if(dxva_result != 0)
					codec_ctx_new->thread_count = inffmpgdec_get_nb_sw_video_decoder_threads(ffmpi, codec_ctx_new);
			}
			mpxplay_debugf(MPXPLAY_DEBUG_OPENCLSDTL, "in_ffmpgdec_infile_callback avcodec_open2 BEGIN ffmpi:%8.8X ctx:%8.8X cav:%8.8X i:%d dxva:%d w:%d h:%d",
					(unsigned int)ffmpi, (unsigned int)codec_ctx_new, (unsigned int)codec_av_new, stream_type, dxva_result, codec_ctx_new->coded_width, codec_ctx_new->coded_height);
			if(avcodec_open2(codec_ctx_new, codec_av_new, NULL) < 0)
			{
				mpxplay_debugf(MPXPLAY_DEBUG_ERROR, "in_ffmpgdec_infile_callback avcodec_open2 FAILED! ffmpi:%8.8X ctx:%8.8X cav:%8.8X i:%d",
					(unsigned int)ffmpi, (unsigned int)codec_ctx_new, (unsigned int)codec_av_new, stream_type);
				return retcode;
			}
			mpxplay_debugf(MPXPLAY_DEBUG_OPENCLSDTL, "in_ffmpgdec_infile_callback avcodec_open2 DONE ffmpi:%8.8X ctx:%8.8X cav:%8.8X i:%d dxva:%d w:%d h:%d",
					(unsigned int)ffmpi, (unsigned int)codec_ctx_new, (unsigned int)codec_av_new, stream_type, dxva_result, codec_ctx_new->coded_width, codec_ctx_new->coded_height);
		}
		if(ffmpi->flags & INFFMPG_FLAG_STILLPICTURE)
			funcbit_enable(pktlist_elem->flags, MPXPLAY_PACKETLISTFLAG_STILLPICT);
//		if(pktlist_elem->flags & (MPXPLAY_PACKETLISTFLAG_INIT | MPXPLAY_PACKETLISTFLAG_CLEARBUF)) // TODO: INIT ???
		if(pktlist_elem->flags & MPXPLAY_PACKETLISTFLAG_CLEARBUF)
		{
			mpxplay_debugf(MPXPLAY_DEBUG_CALLBACK, "in_ffmpgdec_infile_callback mpxplay_ffmpgdec_packetqueue_clear BEGIN ffmpi:%8.8X i:%d", (unsigned int)ffmpi, stream_type);
			mpxplay_ffmpgdec_packetqueue_clear(&ffmpi->packet_queues[stream_type], MPXPLAY_PACKETLISTFLAG_CLEARBUF);
			mpxplay_debugf(MPXPLAY_DEBUG_CALLBACK, "in_ffmpgdec_infile_callback mpxplay_ffmpgdec_packetqueue_clear END   ffmpi:%8.8X", (unsigned int)ffmpi);
		}
		funcbit_disable(pktlist_elem->flags, MPXPLAY_PACKETLISTFLAG_INIT);
		*packet_dest = pktlist_elem;
		retcode = MPXPLAY_ERROR_INFILE_OK;
	}

	return retcode;
}

int in_ffmpgdec_infile_callback(void *demuxer_data, unsigned int control, mpxp_ptrsize_t arg1, mpxp_ptrsize_t arg2, int timeout_ms)
{
	struct ffmpg_demuxer_data_s *ffmpi = demuxer_data;
	int retcode = MPXPLAY_ERROR_INFILE_NODATA;
	mpxplay_packetlist_t *pktlist_elem;

	if(!ffmpi)
		return retcode;

	if(timeout_ms != INFFMPG_MUTEXTIME_NOLOCK)
	{
		retcode = PDS_THREADS_MUTEX_LOCK(&ffmpi->mutexhnd_ffmpeg, timeout_ms);
		if(retcode != MPXPLAY_ERROR_OK){
			if(timeout_ms > MPXPLAY_MUTEXTIME_SHORT){
				mpxplay_debugf(MPXPLAY_DEBUG_WARNING, "in_ffmpgdec_infile_callback pds_threads_mutex_lock FAILED cmd:%8.8X pd:%8.8X ret:%d tms:%d", control, (int)demuxer_data, retcode, timeout_ms);
			}
			return retcode;
		}
	}

	switch(control)
	{
		case MPXPLAY_INFILE_CBKCTRL_SET_SELECT_PROGRAM:
			ffmpi->select_new_program_number = arg1;
			retcode = MPXPLAY_ERROR_INFILE_OK;
			mpxplay_debugf(MPXPLAY_DEBUG_PRGCHG, "in_ffmpgdec_infile_callback SELECT_PROGRAM %d", ffmpi->select_new_program_number);
			break;
		case MPXPLAY_INFILE_CBKCTRL_SET_SELECT_STREAM:
			//mpxplay_debugf(MPXPLAY_DEBUG_CALLBACK, "MPXPLAY_INFILE_CBKCTRL_SET_SELECT_STREAM %d -> %d", arg1_num, *arg2);
			if(arg1 < MPXPLAY_STREAMTYPEINDEX_NUM){
				ffmpi->select_new_stream_number[arg1] = arg2;
				retcode = MPXPLAY_ERROR_INFILE_OK;
			}
			break;
		case MPXPLAY_INFILE_CBKCTRL_SET_PACKETQUEUE_CLEAR:
			if(arg1 < MPXPLAY_STREAMTYPEINDEX_NUM)
				retcode = mpxplay_ffmpgdec_packetqueue_clear(&ffmpi->packet_queues[arg1], arg2);
			break;
		case MPXPLAY_INFILE_CBKCTRL_SET_CODECCTX_CLEAR:
			if(arg1 < MPXPLAY_STREAMTYPEINDEX_NUM)
				retcode = mpxplay_ffmpgdec_packetqueue_clear(&ffmpi->codecctx_queues[arg1], arg2);
			break;
		default:
			if(!arg1) // used as a pointer bellow
				goto err_out_cbk;
	}

	switch(control){
#ifdef MPXPLAY_GUI_QT
		case MPXPLAY_INFILE_CBKCTRL_SET_ADDSTREAMFILESTOVIDEO:
			retcode = in_ffmpfile_filelist_files_add_lfcbds(ffmpi, (struct mpxplay_loadfile_callbackdata_s *)arg1);
			break;
		case MPXPLAY_INFILE_CBKCTRL_GET_PROGRAMLIST:
			*((struct mpxplay_programinfo_t **)arg1) = inffmpg_programinfolist_fill(ffmpi);
			retcode = MPXPLAY_ERROR_INFILE_OK;
			break;
		case MPXPLAY_INFILE_CBKCTRL_GET_STREAMLIST:
			*((struct mpxplay_streaminfo_t **)arg1) = inffmpg_streaminfolist_fill(ffmpi);
			retcode = MPXPLAY_ERROR_INFILE_OK;
			break;
#endif // MPXPLAY_GUI_QT
		case MPXPLAY_INFILE_CBKCTRL_GET_TIMESYNC:
			retcode = inffmpgdec_get_avtimesync_us(ffmpi, arg2, (mpxp_int64_t *)arg1);
			break;
		case MPXPLAY_INFILE_CBKCTRL_GET_VIDEOTIMESTAMP:
			*((mpxp_int64_t *)arg1) = ffmpi->avsync_videotimestamp;
			mpxplay_debugf(MPXPLAY_DEBUG_TIMESTAMP, "in_ffmpgdec_infile_callback GET VIDEOTIMESTAMP: %llu", ffmpi->avsync_videotimestamp);
			retcode = MPXPLAY_ERROR_INFILE_OK;
			break;
		case MPXPLAY_INFILE_CBKCTRL_SET_VIDEOTIMESTAMP:
			ffmpi->avsync_videotimestamp = inffmpg_get_avframe_pts_corrected(ffmpi, MPXPLAY_STREAMTYPEINDEX_VIDEO, (AVFrame *)arg1);
			mpxplay_debugf(MPXPLAY_DEBUG_TIMESTAMP, "in_ffmpgdec_infile_callback SET VIDEOTIMESTAMP: %llu", ffmpi->avsync_videotimestamp);
			retcode = MPXPLAY_ERROR_INFILE_OK;
			break;
#ifdef MPXPLAY_GUI_QT
		case MPXPLAY_INFILE_CBKCTRL_SET_VIDFRAME_METADATA:
			inffmpg_pictureinfo_metadata_update(ffmpi, (AVDictionary *)arg1);
			retcode = MPXPLAY_ERROR_INFILE_OK;
			break;
#endif // MPXPLAY_GUI_QT
		case MPXPLAY_INFILE_CBKCTRL_GET_CODECCTXINIT_FFMPEG:
			retcode = inffmpgdec_codecctx_init(ffmpi, arg2, (mpxplay_packetlist_t **)arg1);
			break;
		case MPXPLAY_INFILE_CBKCTRL_GET_CODECCTX_FFMPEG:
			if(arg2 < MPXPLAY_STREAMTYPEINDEX_NUM){
				mpxplay_debugf(MPXPLAY_DEBUG_CALLBACK, "in_ffmpgdec_infile_callback MPXPLAY_INFILE_CBKCTRL_GET_CODECCTX_FFMPEG ffmpi:%8.8X sti:%d", (unsigned int)ffmpi, arg2);
				pktlist_elem = NULL;
				mpxplay_ffmpgdec_packetqueue_get((void *)ffmpi, &ffmpi->codecctx_queues[arg2], &pktlist_elem, 0);
				if(pktlist_elem && pktlist_elem->codec_ctx){
					*((mpxplay_packetlist_t **)arg1) = pktlist_elem;
					retcode = MPXPLAY_ERROR_INFILE_OK;
				}
			}
			break;
		case MPXPLAY_INFILE_CBKCTRL_GET_PACKETLIST_ELEM:
			if(arg2 < MPXPLAY_STREAMTYPEINDEX_NUM){
				pktlist_elem = NULL;
				retcode = mpxplay_ffmpgdec_packetqueue_get((void *)ffmpi, &ffmpi->packet_queues[arg2], &pktlist_elem, 0);
				if(pktlist_elem && (pktlist_elem->flags & MPXPLAY_PACKETLISTFLAG_CLEARBUF) && pktlist_elem->codec_ctx){
					AVCodecContext *packet_ctx = (AVCodecContext *)pktlist_elem->codec_ctx;
					if(avcodec_is_open(packet_ctx) && packet_ctx->codec)
						avcodec_flush_buffers(packet_ctx);  // TODO: right place?
				}
				*((mpxplay_packetlist_t **)arg1) = pktlist_elem;
				retcode = MPXPLAY_ERROR_INFILE_OK;
			}
			break;
		case MPXPLAY_INFILE_CBKCTRL_SET_CODECCTX_CONSOLIDATE:
			if(arg2 < MPXPLAY_STREAMTYPEINDEX_NUM)
				retcode = mpxplay_ffmpgdec_queue_clear_to_elemctx(&ffmpi->codecctx_queues[arg2], (void *)arg1);
			break;
		default:
			if(!arg2)  // other controls need 2 pointer arguments
				goto err_out_cbk;
	}

	switch(control)
	{
		case MPXPLAY_INFILE_CBKCTRL_GET_CORRECTEDVTIMESTAMP:
			//ffmpi->stream_last_pts[MPXPLAY_STREAMTYPEINDEX_VIDEO] = INFFMPG_INVALID_PTS;
			*((mpxp_int64_t *)arg2) = inffmpg_get_avframe_pts_corrected(ffmpi, MPXPLAY_STREAMTYPEINDEX_VIDEO, ((AVFrame *)arg1));
			//if(first_pktlist_elem->flags & MPXPLAY_PACKETLISTFLAG_CLEARBUF)
			{
				mpxplay_debugf(MPXPLAY_DEBUG_TIMESTAMP,"CBKCTRL_GET_CORRECTEDVTIMESTAMP us:%llu diff:%lld pts:%lld ppts:%lld pdts:%lld cpn:%d ts:%lld",
					*((mpxp_int64_t *)arg2), (*((mpxp_int64_t *)arg2) - ffmpi->stream_last_pts[MPXPLAY_STREAMTYPEINDEX_VIDEO]),
					((AVFrame *)arg1)->pts, ((AVFrame *)arg1)->pkt_pts, ((AVFrame *)arg1)->pkt_dts, ((AVFrame *)arg1)->coded_picture_number,
					(pds_gettimeu() - ffmpi->avsync_clocktime + ffmpi->avsync_timestamp));
			}
			ffmpi->stream_last_pts[MPXPLAY_STREAMTYPEINDEX_VIDEO] = *((mpxp_int64_t *)arg2);
			retcode = MPXPLAY_ERROR_INFILE_OK;
			break;
	}

err_out_cbk:
	if(timeout_ms != INFFMPG_MUTEXTIME_NOLOCK)
	{
		PDS_THREADS_MUTEX_UNLOCK(&ffmpi->mutexhnd_ffmpeg);
	}

	return retcode;
}

#endif //MPXPLAY_LINK_INFILE_FF_MPEG
